<?php
/*
 * @author Ralph Bolton, July 2004, www.coofercat.com
 * @version 1.0
 * @copyright Copyright &copy; 2004 Ralph Bolton.
 * @license Tiki License
 * @see http://tikiwiki.org for Licensing information.
*/

/*
 * Outbound URL Translation (URLt) Smarty Output Filter
 *
 * This filter "reads" the HTML about to go to a client and
 * translates the URLs in it to a different form (defined by
 * the administrator in a database table).
 *
 * This filter uses a (fast and dirty) HTML parser. Proper parsers
 * such as the one at http://php-html.sourceforge.net/) take far
 * too long parsing all the tags in the HTML. Since URLt only
 * requires a very small subset of tags, the necessary funtionality
 * is realised in Perl regular expressions (regexes).
 *
 * In general, HTML parsing by regex is dangerous, because most
 * browsers will render very messy looking HTML which regex writers
 * are not expecting. The regexes in this filter are designed to be
 * forgiving of messy HTML, although changes are expected in this
 * area as we find the parser breaks on some obscure HTML page.
 *
 * PHP developers looking for an HTML parser should use extreme
 * caution before stealing this code. Firstly, I wrote it, so it's
 * probably not that reliable (!) and secondly, it's designed to
 * work with Tiki (1.8.3) which is supposed to only output HTML
 * Standards compliant HTML. As a result, it almost certainly will
 * not work with less compliant HTML.
 *
 * That said, this filter can be used to modify Tiki URLs before
 * they are seen by the user's browser. Thus, the links can be
 * made to work in a different scheme from the default, and can
 * be made to be more user-friendly and search engine compatible.
 *
 * For example, a Wiki page would normally have a URL something
 * like: http://www.coofercat.com/tiki-1.8.3/tiki-index.php?page=CooferCatWiki
 * After translation, it could easily be:
 * http://www.coofercat.com/wiki/CooferCatWiki
 * ...which is both search engine friendly, and very memorable for
 * humans. It also hides the fact the site is using Tiki to provide
 * Wiki features (thus can be de-branded).
 *
 * See the (proper) documentation for details of how outbound URL
 * translation can be used (and what you need to do to get the
 * inbound URLs to actually return pages ;-)
*/

// Initialisation
// Note: Things like "require_once tiki-setup.php" are NOT
// required, as they're already in-scope. Just use the global
// $tikilib at will ;-)


/*
 * string urlt_perform_regexes(string url)
 * This function takes a URL (possibly relative or fully qualified,
 * of any scheme) and runs a series of configurable regexes on it.
 * The regexes have a flag associated with them that can cut short
 * processing. That is, stop processing any more regexes if a match
 * was found or not found, for example. What ever result is obtained
 * is returned.
 *
 * Regexes are found in the tiki_urlt_regex_out database table.
 *
 * @access private
*/
function urlt_perform_regexes($url)
{
	//error_log("URL to transform is $url");

	global $tikilib;

	// ADOdb style DB access.
	// We use a short cache here so that we keep the table in cache between
	// calls to the output filter (which are plentiful with each Tiki request).
	// There may be good reason to have a long cache here, at the cost of
	// reaction-to-configuration times. 
	if(($rs=$tikilib->db->CacheExecute(1,"select * from tiki_urlt_regex_out"))===false)
	{
		// Failed to read the database properly.
		error_log("urlt_out: Failed to Read database: " . $tikilib->db->ErrorMsg);
		return $url;
	}
	while ($array = $rs->FetchRow())
	{
		//error_log("ID: ".$array['regexid']." Left: ".$array['regex_left']." Right: ".$array['regex_right']." Flags: ".$array['regex_flags']." Continue: ".$array['continue_flag']);

		//$newurl=preg_replace('/'.$array['regex_left'].'/'.$array['regex_flags'], $array['regex_right'], $url);

		// First, see if the rule is disabled - if so, skip it.
		if($array['continue_flag'] == 3)
		{
			// Rule disabled - skip it
			//error_log("Skipping disabled rule " . $array['regexid']);
			continue;
		}

		// We have to check first, replace second. We can't just attempt the replace
		// because there's no way to know if it was successful if you don't change
		// anything (which is possible - and likely!)
		//if($newurl == $url)
		if(!preg_match('/'.$array['regex_left'].'/'.$array['regex_flags'],$url))
		{
			# No change - no match!
			if($array['continue_flag'] == 2)
			{
				//error_log("Ceasing processing because flag = 2 (stop on no match)");
				break;
			}
		}
		else
		{
			# Changed - Match
			//$url=$newurl;
			$url=preg_replace('/'.$array['regex_left'].'/'.$array['regex_flags'], $array['regex_right'], $url);
			if($array['continue_flag'] == 1)
			{
				//error_log("Ceasing processing because flag = 1 (stop on match)");
				break;
			}
		}
	}

	//error_log("Transformed URL is $url");
	return $url;
}

/*
 * string urlt_parse_html(string html)
 * This function takes a chunk of HTML and attempts to find URLs in it.
 * The parsing is "fast and dirty", using regular expressions rather
 * than a proper HTML parser. This is purely a speed issue - "real" parsers
 * are far too slow at this work. This solution is considerably faster
 * although prone to flaky HTML. However, since Tiki is supposed to only
 * output standards compliant HTML, flaky HTML is unlikely to be encountered.
 *
 * Using this code for anything other than it's intended purpose may
 * be risky.
 *
 * This function iterates over the HTML it is given, looking for various
 * HTML tags (eg. <a ...> <img ...> <link ...> and <script ...> (etc). When a
 * tag is found, the URL in it (if any) is extracted and passed to
 * urlt_perform_regexes() for transformation. The result is then pasted back
 * into the HTML in place of the original before the HTML is returned to the
 * caller.
 *
 * Although care has been taken to write fast regexes, this is likely to
 * be slow on large files, very complex files etc. 
 *
 * @access private
*/

function urlt_parse_html($html)
{
	$offset=0;
	$matches=array();
	// Match <a ...> <img ...> <link ...> <script ...> and <form ...>
	// Note:Very defensive regexes in use here. Firstly, avoiding the use of case insensitive
	// regex parsing (ie. /i), and also generous allowances for spaces in the tag, so tags like
	// <    a   \n   href  \n =    http://site/"> are allowed, even though they're messy and
	// possibly illegal HTML.
	while(preg_match('/\<\s*(a|img|link|script|form|option)\s+([^\>]*)\>/i',
		$html,$matches,PREG_OFFSET_CAPTURE,$offset))
	{
		$tag_text=$matches[0][0];
		$tag_offset=$matches[0][1];

		// By now, $tag_text contains something like:
		// <a class="link" href="http://www.coofercat.com/" target="_blank">
		// or: <img src="http://www.coofercat.com/img.gif" width="100" height="200"/>
		// (etc)

		//error_log("Tag is $tag_text at offset $tag_offset");

		// Now see if we can get an href="... or a src="... attribute...

		// Need these in a moment...
		$left="";
		$right="";
		$url="";

		// Now handle href= and src= attributes... (assumption: Only one of
		// href or src is allowed in the tag)
		if(preg_match('/(href|src|action|value)[\s\n]*=[\s\n]*[\"\']?/i',
			$tag_text,$temp_matches,PREG_OFFSET_CAPTURE,0))
		{
			// We're now going to do lots of text manipulation. Here's an
			// example:
			// 	href="http://www.coofercat.com/" target="_blank">
			// will become:
			//	Left:	href="
			//	URL:	http://www.coofercat.com/
			//	Right:	" target="_blank">

			// Get the position of the end of the match we just made
			// (start of match + length of match)
			$thing=$temp_matches[0][1] + strlen($temp_matches[0][0]);

			// Snip out the text up to the end of the match we just made...
			$left=substr($tag_text,0,$thing);

			//error_log("Left is $left thing is $thing");

			// Now get the text beyond the match we just made...
			$temp=substr($tag_text,strlen($left));

			// Now get the right hand side...
			// (ie. stuff after the URL)
			// Is this matching optimal? Does it miss things? - Ralph
			if(preg_match('/^[^\"\'\>]*/',$temp,$temp_matches,PREG_OFFSET_CAPTURE,0))
			{
				// Good! We've matched something (which is the URL).
				// Get the length of the match, and snip it out of the
				// text. Then get the bit after that as well.
				$thing=$temp_matches[0][1] + strlen($temp_matches[0][0]);
				$url=substr($temp,0,$thing);
				$right=substr($temp,strlen($url));

				//error_log("url is $url right is $right");
			}
		}
		// Space and scope here for other (more complex) tag types that require
		// more careful handling. Obviously not implemented though!
		
		if($url != "")
		{
			// We've got a URL (ie. http://www.coofercat.com/
			// Transform it...
			$url=urlt_perform_regexes($url);
			// Put it back in the tag...
			// (ie. <a href="http://www.coofercat.com/" target="_blank">)
			$newtag=$left . $url . $right;
			// Now put it back in the text...
			$first=substr($html,0,$tag_offset);
			$second=substr($html,$tag_offset + strlen($tag_text));
			$html=$first . $newtag . $second;
			// Now move the matching offset along just beyond what
			// we've inserted.
			$offset=$tag_offset+strlen($newtag);
		}
		else
		{
			// Otherwise, just move the offset along so that
			// we don't rematch anything we just found.
			$offset=$tag_offset+strlen($tag_text);
		}
	}

	// Return what ever the hell we've managed to come up with ;-)
	return $html;
}

/*
 * string smarty_outputfilter_urlt(string text)
 * This is the main function of the URLt filter. It is called by
 * Smarty to perform output filtering. Smarty expects to be given
 * a filtered string in return for an unfiltered one.
 * @see http://smarty.php.net/manual/en/advanced.features.outputfilters.php
 *
 * This function should check to see if it is required by the administrator
 * by looking to see if it's "feature" is switched on. This is not implemented
 * yet, but should be before proper insertion into the Tiki distribution.
 * (over to real Tiki developers for that bit ;-)
 *
 * Note: it seems that the output filter is called numerous times for each
 * Tiki request. This (I assume) is for each sub-template that is used to build
 * a page. Care should be taken with any HTML work in this regard, as you
 * may not be dealing with a complete and well formed HTML page. Smarty
 * doesn't seem to have any hooks that can be used to detect if this is the
 * last call to the filter for this request, so it's no good trying to
 * bundle up text for processing in one go once you've got it all.
 * Also Note: Smarty seems to re-process templates, such that they are
 * passed here more than once. That more or less negates the benefit of
 * the "stop on match" functionality, and is obviously inefficient.
 *
 * To configure this filter into Tiki/Smarty, edit setup_smarty.php
 * and add the line "$smarty->load_filter('output','urlt');" at the end
 * of the file (you'll find other output filter stuff there). This script
 * should then be called lib/smarty/plugins/outputfilter.urlt.php
 *
 * @access restricted
*/
function smarty_outputfilter_urlt($text)
{
	// Check if URLt is required by the administrator
	// (ie, is feature enabled?)
	// NOT IMPLEMENTED (in Tiki)
	//if ($feature_urlt == 'y') {

		// Perform URLt on a chunk of HTML...
		// Should we pass by reference for speed? - Ralph
		// (my experience is that chunks of HTML are quite small, so "no"! - Ralph)
		$text=urlt_parse_html($text);

		// Now check the headers...
		// NOT IMPLEMENTED - is this even possible with PHP4?
	//}

	// Hand the modified HTML back to Smarty...
	return $text;
}

?>
